/*
 * Generate a VGA signal using a PIC32 microcontroller.
 *
 * Copyright (C) 2017, 2018 Paul Boddie <paul@boddie.org.uk>
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */


#include "pic32_c.h"
#include "init.h"
#include "debug.h"

/* Specific functionality. */

#include "main.h"
#include "devconfig.h"
#include "vga.h"
#include "display.h"
#include "display_config.h"
#include "vga_display.h"



/* Define timers if not indicated in the build configuration. */

#ifndef LINE_CHANNELS
#define LINE_CHANNELS   1
#endif

#ifndef LINE_TIMER
#define LINE_TIMER      2
#endif

#ifndef TRANSFER_TIMER
#define TRANSFER_TIMER  0
#endif

/* Define different output ports and pins for parallel mode. */

#ifndef PARALLEL_MODE
#define VGA_OUTPUT      PORTB
#define OC1_PIN         RPA0R
#define OC2_PIN         RPA1R
#define LED_PIN         (1 << 3)
#else
#define VGA_OUTPUT      PM_REG(0, PMxDIN)
#define OC1_PIN         RPB4R
#define OC2_PIN         RPB5R
#define LED_PIN         (1 << 2)
#endif



/* Bundled image data. */

extern uint8_t screendata[];
extern uint32_t screendata_width, screendata_height;



/* Blink an attached LED with delays implemented using a loop. */

static void blink(uint32_t delay, uint32_t port, uint32_t pins)
{
    uint32_t counter;

    /* Clear outputs (LED). */

    CLR_REG(port, pins);

    while (1)
    {
        counter = delay;

        while (counter--) __asm__("");	/* retain loop */

        /* Invert outputs (LED). */

        INV_REG(port, pins);
    }
}



/* Main program. */

void main(void)
{
    init_memory();
    init_pins();
    init_outputs();

    unlock_config();
    config_oc();
    config_uart();
    lock_config();

    init_dma();

#ifdef PARALLEL_MODE
    init_pm();

    /* Configure parallel master mode. */

    pm_init(0, 0b10);
    pm_set_output(0, 1, 0);
    pm_on(0);
#endif

    /* Initialise VGA output with one or two line channels, configuring a line
       timer and any transfer timer, with an initiating channel being introduced
       if a transfer timer is specified. */

    init_vga_with_timers(&display_config, LINE_CHANNELS, LINE_TIMER, TRANSFER_TIMER);

    /* Configure VGA output transfer to the output register, also configuring
       output compare units for horizontal and vertical sync. */

    vga_configure_transfer(VGA_OUTPUT);
    vga_configure_sync(1, 2);

    uart_init(1, FPB, 115200);
    uart_on(1);

    interrupts_on();

    copy_to_framebuffer(&display_config, screendata, screendata_width, screendata_height);
    blink(1 << 24, PORTA, LED_PIN);
}



/* Exception and interrupt handlers. */

void exception_handler(void)
{
    blink(3 << 12, PORTA, LED_PIN);
}

void interrupt_handler(void)
{
    uint32_t ifs;

    /* Check for a OC1 interrupt condition. */

    ifs = REG(OCIFS) & OC_INT_FLAGS(1, OCxIF);

    if (ifs)
    {
        vga_interrupt_handler();
        CLR_REG(OCIFS, ifs);
    }
}



/* Peripheral pin configuration. */

void config_oc(void)
{
    /* Map OC1. */

    REG(OC1_PIN) = 0b0101;

    /* Map OC2. */

    REG(OC2_PIN) = 0b0101;
}

void config_uart(void)
{
    /* Map U1RX to RPB13. */

    REG(U1RXR) = 0b0011;        /* U1RXR<3:0> = 0011 (RPB13) */

    /* Map U1TX to RPB15. */

    REG(RPB15R) = 0b0001;       /* RPB15R<3:0> = 0001 (U1TX) */

    /* Set RPB13 to input. */

    SET_REG(TRISB, 1 << 13);
}
